//------------------------------------------------------------------------------
// CERN BE-CO-HT
// White Rabbit Trigger Distribution (WRTD)
// https://ohwr.org/projects/wrtd
//------------------------------------------------------------------------------
//
// unit name:   WrtdDrv
//
// description: A SystemVerilog Class to provide an abstraction of a complete
// WRTD device.
//
//------------------------------------------------------------------------------
// Copyright CERN 2019
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __WRTD_DRV_INCLUDED
 `define __WRTD_DRV_INCLUDED

 `include "wrtd_definitions.svh"
 `include "wrtd_rule.svh"
 `include "wrtd_alarm.svh"
 `include "mock_turtle_driver.svh"

class WrtdDrv;
   uint32_t  ready;
   protected string name;
   protected MockTurtleDriver     mt;
   protected uint32_t             nbr_cpus;
   protected uint32_t             free_rule_slots[];
   protected uint32_t             hmq_words[];
   protected wrtd_mt_config       cfg;
   protected wrtd_root            roots[];
   protected wrtd_dev             devices[$];
   protected WrtdRepCapCollection rules;
   protected WrtdRepCapCollection alarms;

   function new ( CBusAccessor acc, uint64_t base,
		  vIMockTurtleIRQ irq, string name = "" );
      this.name = name;
      this.mt   = new (acc, base, irq, name);
   endfunction // new

   task mdisplay ( string str );
      string tmp;
      if (this.name == "")
	tmp = $sformatf("<%t> %s", $realtime, str);
      else
	tmp = $sformatf("[%s] <%t> %s", this.name, $realtime, str);
      $display (tmp);
   endtask // mdisplay

   task init ( );
      int i, j;

      uint32_t val, cpu_ready;

      WrtdRule new_rule;

      WrtdAlarm new_alarm;

      wrtd_dev new_dev;

      uint32_t data[];

      this.ready = 0;

      mt.init ( );

      // TODO: replace with non-hardcoded values
      this.cfg.mt_app_id = 'h115790d1;
      this.cfg.nbr_cpus  = 1;
      this.cfg.tx_cpu    = 0;
      this.cfg.rx_cpu    = 0;

      this.cfg.cpus[0].fw_id       = 'h35b0;
      this.cfg.cpus[0].nbr_devices = 2;

      this.cfg.cpus[0].devices[0].name        = "FMC ADCin 5CH";
      this.cfg.cpus[0].devices[0].dir         = WRTD_DIR_INPUT;
      this.cfg.cpus[0].devices[0].nbr_chs     = 5;
      this.cfg.cpus[0].devices[0].channels    = new[5];
      this.cfg.cpus[0].devices[0].channels[0] = "ADCI1";
      this.cfg.cpus[0].devices[0].channels[1] = "ADCI2";
      this.cfg.cpus[0].devices[0].channels[2] = "ADCI3";
      this.cfg.cpus[0].devices[0].channels[3] = "ADCI4";
      this.cfg.cpus[0].devices[0].channels[4] = "ADCI5";

      this.cfg.cpus[0].devices[1].name        = "FMC ADCout";
      this.cfg.cpus[0].devices[1].dir         = WRTD_DIR_OUTPUT;
      this.cfg.cpus[0].devices[1].nbr_chs     = 1;
      this.cfg.cpus[0].devices[1].channels    = new[1];
      this.cfg.cpus[0].devices[1].channels[0] = "ADCO1";

      this.nbr_cpus = mt.rom.getCoreCount ( );

      this.free_rule_slots = new[this.nbr_cpus];

      this.hmq_words = new[this.nbr_cpus];

      this.roots = new[this.nbr_cpus];

      for ( i = 0; i < this.nbr_cpus; i++ )
	begin
	   this.hmq_words[i] =
		`TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD( mt.rom.getHmqDimensions( i, 0 ) );
	   mt.enable_hmqi_irq ( i, 0, 1 );
	   mt.enable_console_irq ( i, 1 );
	   mt.reset_core ( i, 0 );
	end

      fork
	 begin

	    for ( i = 0; i < this.nbr_cpus; i++ )
	      begin
		 // wait for ready notification from firmware
		 cpu_ready = 0;
		 while ( cpu_ready == 0 )
		   begin
		      val = 0;
		      mt.get_single_cpu_notification ( 0, val );
		      if ( val == TRTL_CPU_NOTIFY_MAIN )
			cpu_ready = 1;
		      # 1us;
		   end
		 // retrieve address of root
		 msg_get_config ( i, data );
		 this.roots[i].addr = data[0];
		 // retrieve root
		 msg_readw ( i, this.roots[i].addr, `WRTD_ROOT_WORD_SIZE, data );
		 this.roots[i].nbr_rules      = ( data[1] & 'h00ff0000 ) >> 16;
		 this.roots[i].nbr_alarms     = ( data[1] & 'h0000ff00 ) >> 8;
		 this.roots[i].nbr_devs       = ( data[1] & 'h000000ff ) >> 0;
		 this.roots[i].devs_addr[0:3] = data[4:7];
		 this.roots[i].rules_addr     = data[8];
		 this.roots[i].alarms_addr    = data[9];
		 // init free rule slots
		 this.free_rule_slots[i] = this.roots[i].nbr_rules;
	      end

	    // initialise rules
	    this.rules  = new ( this.name );
	    for ( i = 0; i < this.nbr_cpus; i++ )
	      for ( j = 0; j < this.roots[i].nbr_rules; j++ )
		begin
		   new_rule = new ( i, j, this.name );
		   this.rules.collection.push_back ( new_rule );
		end

	    // initialise alarms
	    this.alarms = new ( this.name );
	    for ( i = 0; i < this.nbr_cpus; i++ )
	      for ( j = 0; j < this.roots[i].nbr_alarms; j++ )
		begin
		   new_alarm = new ( i, j, this.name );
		   this.alarms.collection.push_back ( new_alarm );
		end

	    // initialise devices
	    this.devices.delete();
	    for ( i = 0; i < this.nbr_cpus; i++ )
	      for ( j = 0; j < this.roots[i].nbr_devs; j++ )
		begin
		   new_dev.core  = i;
		   new_dev.index = j;
		   new_dev.ch_en = 0;
		   this.devices.push_back ( new_dev );
		end

	    // TODO remove hard-coded ADC init
	    adc_init ();

	    this.ready = 1;
	 end

	 begin
	    while ( this.ready == 0 )
	      begin
		 mt.update ( );
		 # 1us;
	      end
	 end
      join

   endtask // init

   task msg_get_config ( int core, ref uint32_t data[] );
      int i;

      MQueueMsg msg;

      if ( data.size() < `WRTD_CFG_MSG_WORD_SIZE )
	data = new[`WRTD_CFG_MSG_WORD_SIZE];

      msg = new ( core, 0 );
      msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
      msg.header.msg_id = WRTD_ACTION_GET_CONFIG;
      msg.header.len    = 0;
      mt.hmq_send_message ( msg );
      while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
      mt.hmq_receive_message ( msg );
      if ( msg.header.len != `WRTD_CFG_MSG_WORD_SIZE )
	begin
	   $error ( "get_config: unexpected message header length" );
	   mdisplay ( $sformatf ( "message received: %s", msg.tostring() ) );
	end
      for ( i = 0; i < `WRTD_CFG_MSG_WORD_SIZE; i++ )
	data[i] = msg.data[i];

   endtask // msg_get_config

   task msg_readw ( int core, uint32_t addr,
		    uint32_t count, ref uint32_t data[] );
      int i;
      MQueueMsg msg;
      uint32_t tlen, offset;

      if ( data.size() < count )
	data = new[count];

      offset = 0;

      while ( offset < count )
	begin
	   tlen = this.hmq_words[core] - `WRTD_IO_MSG_WORD_SIZE;
	   if ( tlen > count - offset )
	     tlen = count - offset;
	   msg = new ( core, 0 );
	   msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
	   msg.header.msg_id = WRTD_ACTION_READW;
	   msg.header.len    = `WRTD_IO_MSG_WORD_SIZE;
	   msg.data          = '{ addr + offset, tlen };
	   mt.hmq_send_message ( msg );
	   while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
	   mt.hmq_receive_message ( msg );
	   if ( msg.header.len != tlen )
	     begin
		$error ( "readw: unexpected message header length" );
		mdisplay ( $sformatf ( "message received: %s", msg.tostring() ) );
	     end
	   for ( i = 0; i < tlen; i++ )
	     data[offset+i] = msg.data[i];
	   offset += tlen;
	end

   endtask // msg_readw

   task msg_writew ( int core, uint32_t addr,
		     uint32_t count, uint32_t data[] );
      int i;
      MQueueMsg msg;
      uint32_t tlen, offset;

      offset = 0;

      while ( offset < count )
	begin
	   tlen = this.hmq_words[core] - `WRTD_IO_MSG_WORD_SIZE;
	   if ( tlen > count - offset )
	     tlen = count - offset;
	   msg = new ( core, 0 );
	   msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
	   msg.header.msg_id = WRTD_ACTION_WRITEW;
	   msg.header.len    = `WRTD_IO_MSG_WORD_SIZE + count;
	   msg.data          = '{ addr + offset, tlen };
	   for ( i = 0; i < tlen; i++ )
	     msg.data[i+2] = data[offset+i];
	   mt.hmq_send_message ( msg );
	   while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
	   mt.hmq_receive_message ( msg );
	   offset += tlen;
	end

   endtask // msg_writew

   task adc_init ( );
      uint32_t addr;
      uint32_t data[1];

      addr    = this.roots[0].devs_addr[0] + 'h4;
      data[0] = 'h10f;

      msg_writew ( 0, addr, 1, data );
   endtask // adc_init

   task add_rule ( string rep_cap_id );
      int idx;
      idx = this.rules.add ( rep_cap_id );
   endtask // add_rule

   function int find_rule ( string rep_cap_id );
      return this.rules.find ( rep_cap_id );
   endfunction // find_rule

   task remove_rule ( string rep_cap_id );
      int idx;
      idx = this.rules.remove ( rep_cap_id );
      write_rule ( idx );
   endtask // remove_rule

   task set_rule ( string rep_cap_id,
		   string src, string dst,
		   uint32_t delay_ns );
      int idx;

      WrtdRule rule;

      idx = this.rules.find ( rep_cap_id );

      if ( idx == -1 )
	$error ( "%s repeated capability ID cannot be set because it does not exist", rep_cap_id );

      if ( this.rules.collection[idx].is_enabled ( ) )
	$error ( "%s repeated capability ID cannot be set because it is enabled", rep_cap_id );

      $cast ( rule, this.rules.collection[idx] );

      rule.set_src ( src );
      rule.set_dst ( dst );
      rule.set_delay_ns ( delay_ns );
   endtask // set_rule

   task write_rule ( int idx );
      int core, hash;
      uint32_t addr;
      uint32_t data[];

      WrtdRule rule;

      if ( idx < 0 )
	$error ( "cannot write rule with negative index" );

      $cast ( rule, this.rules.collection[idx] );

      hash = wrtd_gen_hash ( rule.get_src() ) % this.rules.collection.size();
      core = rule.get_core ( );
      addr = this.roots[core].rules_addr + hash * `WRTD_RULE_WORD_SIZE * 4;
      data = rule.data_pack ( );

      msg_writew ( core, addr, `WRTD_RULE_WORD_SIZE, data );
   endtask // write_rule

   task enable_rule ( string rep_cap_id );
      int idx;
      string id;
      uint32_t src_cpu;

      WrtdRule rule;

      idx = this.rules.find ( rep_cap_id );

      if ( idx == -1 )
	$error ( "%s repeated capability ID cannot be enabled because it does not exist", rep_cap_id );

      $cast ( rule, this.rules.collection[idx] );

      rule.set_enable ( );

      // TODO, replace hard-coding
      src_cpu = 0;
      if ( this.free_rule_slots[src_cpu] == 0 )
	$error ( "no more available free rule slots on CPU %d", src_cpu );
      this.free_rule_slots[src_cpu] -= 1;
      rule.set_dest_cpu ( `WRTD_DEST_CPU_LOCAL );
      id = rule.get_dst ( );
      if ( id == "ADCO1" )
	rule.set_dest_ch ( 0 );
      else
	rule.set_dest_ch ( `WRTD_DEST_CH_NET );
      write_rule ( idx );
   endtask // enable_rule

   task add_alarm ( string rep_cap_id );
      int idx;
      idx = this.alarms.add ( rep_cap_id );
      write_alarm ( idx );
   endtask // add_alarm

   function int find_alarm ( string rep_cap_id );
      return this.alarms.find ( rep_cap_id );
   endfunction // find_alarm

   task remove_alarm ( string rep_cap_id );
      int idx;
      idx = this.alarms.remove ( rep_cap_id );
      write_alarm ( idx );
   endtask // remove_alarm

   task write_alarm ( int idx );
      int core, index;
      uint32_t addr;
      uint32_t data[];

      if ( idx < 0 )
	$error ( "cannot write alarm with negative index" );

      core  = this.alarms.collection[idx].get_core ( );
      index = this.alarms.collection[idx].get_index ();
      addr  = this.roots[core].alarms_addr + index * `WRTD_ALRM_WORD_SIZE * 4;
      data  = this.alarms.collection[idx].data_pack( );

      msg_writew ( core, addr, `WRTD_ALRM_WORD_SIZE, data );
   endtask // write_alarm

   task log_read ( );
   endtask // log_read

   task update ( );
      mt.update ( );
      fork
/* -----\/----- EXCLUDED -----\/-----
	 check_log_queue ( 0 );
	 check_log_queue ( 1 );
 -----/\----- EXCLUDED -----/\----- */
      join
   endtask // update

endclass // WrtdDrv

`endif //  `ifndef __WRTD_DRV_INCLUDED
